# The following code is confidential and may not be distributed/used without permission from the author.
__author__ = 'Ole Mikkel Sjoelie'

############################
# Imports start
############################

import sys
import random
import math
import copy
import json
import time
import heapq  # For the min-heap in the astar algorithm
import pygame
from pygame.constants import MOUSEBUTTONDOWN
import psycopg2
from ezscroll import N, W, ScrollPane


############################
# Imports end
############################

############################
# Config variables START   #
############################

# The size a tile/node will take on the map.
# The smaller the longer time to generate a path, but it may give more accurate results with smaller tiles.
tileSize = 4

# AStar timeout, the maximum amount of time in seconds that astar is allowed to use per path.
astar_timeout = 20

# Campus id
campusId = 1

# Building information we want to check
#buildingId = 15  # It-syd
#buildingId = 39  # it-bygget
#buildingId = 67  # realfag
buildingId = 33  # Kjemi ?
#buildingId = 2  # hovedbygg
#buildingId = 17  # error-building, works with magnification = 11.

floorValue = 1  # The building floor, 1 is the first floor.

# How much to zoom in
magnification = 12

# doorExpandRadius tells us how much of a radius we should make around doors
# doorExpandRadius = 12
doorExpandRadius = int(1.1 * magnification)


# Distance away from black areas
distanceToCheck = 0

# This is the check distance to the nearest black area, we will favor paths further away from black areas (Middle of rooms etc.)
# This -should- give more consistent paths
costDistanceCheck = 30

#Show AStar colors for openset and closedset
showAStarColors = False

# only generate one path?
generateOnePath = False


# Do we want to run A*, or shall we skip it and just generate the import?
runSearch = True

# If set to True, manually add doors by clicking where they should be before AStar will be run.
manualDoors = False

# Wallcost amount
wallCostAmount = 6

# The distance between the points to check when we try to generate straight lines.
# It's rarely necessary to get the absolute closest point when generating lines, this creates some air and increases performance.
# This should be a value at 1 or higher. If it's set too high, only the start and end point of the line will be considered.
distanceBetweenPoints = 25


############################
#Config END
############################

############################
# Global variables START   #
############################

poiIds = set()

tiles = set()

# Our list of tiles which will be accessible by doing newTiles[x][y] and give you the tile/square at this position.
newTiles = []

# The ID we give to the first tile. This is used to hash each object later on
globalTileId = 1

# For line hashes
globalLineId = 1

# The time when we start the application, so we know how long things take, approximately.
startTime = time.clock()

# The list used for room polygons.
myList = []


# Used for calculating our offset.
lowestLon = None
lowestLat = None

# Just some mouse debug
# Should be used to make paths between rooms etc by clicking on the GUI.
hasPlacedStart = None
hasPlacedEnd = None

# Closest tile to mouse click
mouseClickStart = None
mouseClickGoal = None

# Screen Buffers
screenBuf = None


# All paths added to one place
allPaths = set()

# the size of a tile, for use with newTiles[x][y] for example. the distance between each tile should be curTileSize.
curTileSize = tileSize + 1

# Reduced tiles
reducedHorizontalTiles = set()  # <--->
reducedVerticalTiles = set()  # |
reducedDiagonalTiles1 = set()  # /
reducedDiagonalTiles2 = set()  # \

# QGIS stuff
iface = None


# Lines

# A collection of all our lines
lines = set()

# The line we're most likely going to split
lastLine = None

# Doors
doorPolys = []

#DB Connection
dbUser = "username"
dbName = "database"
dbPwd = "password"

dbConn = psycopg2.connect(database=dbName, user=dbUser, password=dbPwd)
dbCursor = dbConn.cursor()

shouldExitOnCompletion = 0

# Scrollbars

ScrSize = (1000, 800)
Origin = (0, 0)
Gray = (200, 200, 200)
scrollBarSize = 20 # Just the width or height of a scrollbar (width or height depending if its vertical or horizontal)


# Modified lists
# Here we add points or lines that have been modified, so we later can update the database accordingly.
modifiedPoints = set()
modifiedLines = set()



############################
# Global variables END     #
############################



##################################
# Command line arguments START   #
##################################

# Check if we have any command line arguments

if len(sys.argv) >= 3:
    print str(sys.argv)

    # The first argument is usually the script name, for example being called as "python script-name.py <buildingId> <floorValue>"

    # Second argument is the building id taken from the database
    if sys.argv[1] and sys.argv[1].isdigit():
        buildingId = int(sys.argv[1])

    # Third argument is the floor
    if sys.argv[2]:
        floorValue = sys.argv[2]

    if len(sys.argv) >= 4:
        # If not 0, lets exit when the algorithm has been completed
        shouldExitOnCompletion = int(sys.argv[3])

print "Starting import of building " + str(buildingId) + " on floor " + str(floorValue) + "."

##################################
# Command line arguments END   #
##################################

# Point class
# Used to store coordinates, no limitation on what kind of coordinates, what ever fits
# x = longitude
# y = latitude
# origLon and origLat is used to store existing coordinates from a database
# dbId is points taken from the database id.
class Point:
    x = None
    y = None
    origLon = None
    origLat = None
    dbId = None
    isPoi = None

    def __init__(self, xx, yy):
        self.x = xx
        self.y = yy

    def __hash__(self):
        return hash(str(self.x) + " " + str(self.y))


# Line class used for connecting two points
# point1 and point2 are supposed to be of the Point-class
# lineId is used to get a lines id from the database, so we can keep existing lines when we split them etc.
class Line:
    lineId = 0
    dbId = None  # The id of a path in the database.
    point1 = None
    point2 = None

    def __init__(self, p1, p2):
        self.point1 = p1
        self.point2 = p2

        # These IDs are used for hashing, so that each line has an unique ID.
        # That way we can use sets to speed up a lot of calculation, compared to normal python lists.
        global globalLineId
        self.lineId = globalLineId
        globalLineId += 1

    def __hash__(self):
        return self.lineId

    def toString(self):
        return str(self.point1.x) + " " + str(self.point1.y) + ", " + str(self.point2.x) + " " + str(self.point2.y)


# Tile class
# This is the Squares being talked about in the thesis.
# It can be thought of as where we save squares/nodes/tiles on a valid spot in a floor plan.
class Tile:
    x = 0  # local lon-coordinate
    y = 0  # local lat-coordinate
    cost = 10  # Cost of tile
    wallCosts = 0  # Wall cost of tile
    timesUsed = 1  # How many paths using this tile
    origLon = None  # If a tile is linked to an original POI, then we store it here
    origLat = None  # If a tile is linked to an original POI, then we store it here
    dbId = None
    isPoi = None

    def __init__(self, x, y):
        global globalTileId
        self.tileId = globalTileId
        # Increment it for the next tile
        globalTileId += 1
        self.x = x
        self.y = y
        self.curCost = None
        self.estimate = None
        self.parent = None
        self.neighbors = None
        self.timesUsed = 1
        try:
            newTiles[x][y] = self
        except:
            print str(x) + " " + str(y)

    def __cmp__(self, other):
        if self.estimate < other.estimate:
            return -1
        elif self.estimate == other.estimate:
            if self.curCost > other.curCost:
                return -1
            elif self.curCost == other.curCost:
                return 0
            else:
                return 1
        else:
            return 1

    def __hash__(self):
        return self.tileId


# Converts a longitude to the local coordinate system
def importLon(lon):
    global lowestLon
    global magnification
    return int(round((lon - lowestLon) * magnification))


# Converts a latitude to the local coordinate system
def importLat(lat):
    global lowestLat
    global magnification
    return int(round((lat - lowestLat) * magnification))


# Converts the local coordinate system longitude back to the original coordinate system
def exportLon(lon):
    return lowestLon + (lon / magnification)


# Converts the local coordinate system latitude back to the original coordinate system
def exportLat(lat):
    return lowestLat + (lat / magnification)


# Calculates the linear distance from one point to another
def linear_distance(x1, y1, x2, y2):
    return math.sqrt(math.pow((x2 - x1), 2) + math.pow((y2 - y1), 2))


# Distance used in the A* algorithm
def distance(x1, y1, x2, y2):
    return 3 * max(abs(x2 - x1), abs(y2 - y1)) + 1 * min(abs(x2 - x1), abs(y2 - y1))


# Using the same distance in the heuristics
def heuristicCost(start, goal):
    return distance(start.x, start.y, goal.x, goal.y)


# Our function that finds neighbour tiles based on a tile (node) being used as the input
def getNeighbours(node):
    # Neighbour nodes are stored in this set
    nodes = set()

    # The next part could be written with fewer lines of code, but we did it like this to easily comment out certain
    # neighbours, for example if we don't want diagonal neighbours, we can disable all the relevant parts with ease.

    # Right
    if node.x + curTileSize < screenWidth and newTiles[node.x + curTileSize][node.y]:
        nodes.add(newTiles[node.x + curTileSize][node.y])

    # Up
    if node.y - curTileSize >= 0 and newTiles[node.x][node.y - curTileSize]:
        nodes.add(newTiles[node.x][node.y - curTileSize])

    # Down
    if node.y + curTileSize < screenHeight and newTiles[node.x][node.y + curTileSize]:
        nodes.add(newTiles[node.x][node.y + curTileSize])

    # Left
    if node.x - curTileSize >= 0 and newTiles[node.x - curTileSize][node.y]:
        nodes.add(newTiles[node.x - curTileSize][node.y])

    # up right
    if node.x + curTileSize < screenWidth and node.y - curTileSize > 0 and newTiles[node.x + curTileSize][
                node.y - curTileSize]:
        nodes.add(newTiles[node.x + curTileSize][node.y - curTileSize])

    # down right
    if node.x + curTileSize < screenWidth and node.y + curTileSize < screenHeight and newTiles[node.x + curTileSize][
                node.y + curTileSize]:
        nodes.add(newTiles[node.x + curTileSize][node.y + curTileSize])

    # Left down
    if node.x - curTileSize >= 0 and node.y + curTileSize < screenHeight and newTiles[node.x - curTileSize][
                node.y + curTileSize]:
        nodes.add(newTiles[node.x - curTileSize][node.y + curTileSize])

    # Left up
    if node.x - curTileSize >= 0 and node.y - curTileSize > 0 and newTiles[node.x - curTileSize][node.y - curTileSize]:
        nodes.add(newTiles[node.x - curTileSize][node.y - curTileSize])

    # Return all the neighbour tiles
    return nodes


# Creates a path based on the every tiles parents, starting from the goal and going towards the start.
def getPath(cur):
    p = [cur]
    while cur.parent is not None:
        cur = cur.parent
        p.append(cur)
    p.reverse()
    return p


# Our A* implementation, needs a start and goal tile to get it going.
def AStar(start, goal):
    closedSet = set()  # Visited tiles
    openSet = set()  # Tiles to be visited
    openHeap = []  # Sorted based on lowest estimate

    # Add the start tile.
    for tile in tiles:
        # Lets reset the estimate and current cost and parent just in case we're gonna use them again later.
        tile.estimate = None  # Estimated cost
        tile.curCost = None  # Current minimal cost
        tile.parent = None  # The Tile we came from, this lets us backtrack from the goal back to the start to get the shortest path

        if tile.x == start.x and tile.y == start.y:
            heapq.heappush(openHeap, tile)
            openSet.add(tile)
            start = tile

        if tile.x == goal.x and tile.y == goal.y:
            goal = tile

        if not tile.neighbors:
            tile.neighbors = getNeighbours(tile)

        newTiles[tile.x][tile.y] = tile

    start.curCost = 0
    start.estimate = start.curCost + heuristicCost(start, goal) + start.cost + start.wallCosts

    # Lets give AStar a max running time so we're not stuck forever, though we should never actually reach this time-limit.
    astar_timer = time.time()

    while openSet:

        curTime = time.time()
        if (curTime - astar_timer) > astar_timeout:
            # Found no path in time
            print "AStar timeout"
            return None

        # Fetch the object
        current = heapq.heappop(openHeap)
        openSet.remove(current)

        # Yay, we found the goal.
        if current == goal:
            if showAStarColors:
                for tile in openSet:
                    # Color them yellow
                    pygame.draw.rect(allScreen, pygame.color.Color("yellow"),
                                     (tile.x - int(tileSize / 2), tile.y - int(tileSize / 2), tileSize, tileSize))

                for tile in closedSet:
                    # Color closed set dark purple/red-ish
                    pygame.draw.rect(allScreen, (102, 0, 51),
                                     (tile.x - int(tileSize / 2), tile.y - int(tileSize / 2), tileSize, tileSize))

            # Returns the path found.
            return getPath(goal)

        closedSet.add(current)

        if current.neighbors:
            for neighbor in current.neighbors:
                if neighbor in closedSet:
                    continue
                tempCost = current.curCost + distance(current.x, current.y, neighbor.x, neighbor.y) + max(1, (neighbor.cost + neighbor.wallCosts) - neighbor.timesUsed)

                if neighbor not in openSet or tempCost <= neighbor.curCost:
                    neighbor.parent = current
                    neighbor.curCost = tempCost
                    neighbor.estimate = neighbor.curCost + heuristicCost(neighbor, goal) + max(1, (neighbor.cost + neighbor.wallCosts) - neighbor.timesUsed)

                    if neighbor not in openSet:
                        heapq.heappush(openHeap, neighbor)
                        openSet.add(neighbor)

    return None


# Get all Points of Interest (POIs) that we will check if we should visit.
def getPOIList():
    global dbCursor  # database connection

    dbCursor.execute(
        "SELECT ST_AsGeoJSON(geometry), node_id FROM poi WHERE z=" + str(floorValue) + " AND building_id = " + str(
            buildingId) + " AND NOT ST_IsEmpty(geometry) AND GeometryType(geometry) = 'POINT'")

    pois = dbCursor.fetchall()
    poiList = []
    for poi in pois:
        jsonStr = json.loads(poi[0])
        # Lets make a point
        item = Point(jsonStr['coordinates'][0], jsonStr['coordinates'][1])

        # Set the db id, so we can find it later
        item.dbId = poi[1]
        item.isPoi = True

        poiList.append(item)

    return poiList


# Door class used to store information about doors, which will later be used to find the direction of doors
class Door:
    def __init__(self, door_ids, room_ids, center_x, center_y, is_ring, has_arc):
        self.door_ids = door_ids  # Should not include curved part of door
        self.room_ids = room_ids
        self.center_x = center_x
        self.center_y = center_y
        self.is_ring = is_ring
        self.has_arc = has_arc

    def center_wkt(self):
        return 'POINT(' + str(self.center_x) + ' ' + str(self.center_y) + ')'


# getDoors(), With help from Aasmund Tokheim in Wireless Trondheim.
# This function attempts to find all the doors we need to connect rooms.
# Warning, some heavy SQL + postgis incoming :)
def getDoors():
    global dbConn
    global dbCursor

    buffersize = 0.1
    walkable_buffer = 0.8
    doorlayers = ('DORER', )  #Must be tuple
    roomlayers = ('ROM', )

    unionsql = "select st_union(st_buffer(wkb_geometry, %s, 'endcap=square join=bevel')) geometry, " \
               + 'building_id, z from floorplan ' \
               + 'where building_id = %s and z=%s and layer_name in (%s) and st_geometrytype(wkb_geometry) ' \
               + "ilike 'st_linestring' group by building_id, z"
    dumpsql = 'select (st_dump(geometry)).geom doorpoly, building_id, z from (' + unionsql + ') t'
    roomjoin = 'select doorpoly, d.building_id, d.z, array_agg(ogc_fid) room_ids from (' \
               + dumpsql + ') d left outer join floorplan f on d.building_id = f.building_id and d.z = f.z ' \
               + 'and doorpoly && wkb_geometry and st_intersects(doorpoly, wkb_geometry) ' \
               + 'where layer_name in (%s) group by doorpoly, d.building_id, d.z'
    doorjoin = 'select st_x(st_centroid(doorpoly)), st_y(st_centroid(doorpoly)), room_ids, ' \
               + 'array_agg(ogc_fid) door_ids, st_nrings(doorpoly) > 1 as is_ring, max(st_npoints(wkb_geometry)) > 15 as has_arc ' \
               + 'from (' + roomjoin + ') t left outer join floorplan f on ' \
               + 'f.building_id = t.building_id and f.z = t.z and doorpoly && wkb_geometry ' \
               + "and st_contains(doorpoly, wkb_geometry) where layer_name in %s and st_geometrytype(wkb_geometry) ilike 'st_linestring'" \
               + 'and st_length(wkb_geometry) > 0.5 group by doorpoly, room_ids'
    # print doorjoin
    data = (buffersize, buildingId, floorValue, doorlayers, roomlayers, doorlayers)
    dbCursor.execute(doorjoin, data)
    results = dbCursor.fetchall()
    print 'doors fetched: ' + str(len(results))
    doors = []
    for result in results:
        doors.append(Door(result[3], result[2], result[0], result[1], result[4], result[5]))

    for door in doors:
        poly = getDoorDirection(door, buildingId, floorValue, dbCursor, walkable_buffer)
        if poly:
            #print (poly[0])[0]
            jsonStr = json.loads((poly[0])[0])
            coords = jsonStr['coordinates'][0]
            doorPolys.append(coords)


# Using postgis to find the direction of the doors
# Warning, some heavy SQL incoming :)
def getDoorDirection(door, building_id, z, cursor, walkable_buffer):
    dumpdoorlinesegs = 'select st_makeline(st_pointN(wkb_geometry, n), st_pointN(wkb_geometry, n+1)) door ' \
                       + 'from floorplan cross join generate_series(1, 15) n where ' \
                       + 'n < st_npoints(wkb_geometry) and ogc_fid in %s and st_npoints(wkb_geometry) < 15'

    if len(door.door_ids) == 0:
        print 'missing data at ' + door.center_wkt()  # Should not occur
    elif len(door.room_ids) == 0 or not door.has_arc:  # perhaps it is "skyvedoer", just buffer center
        insertsql = "select ST_AsGeoJSON(st_buffer(st_pointfromtext(%s, 900913), %s, 'endcap=square join=bevel'))"
        data = (door.center_wkt(), walkable_buffer)
        cursor.execute(insertsql, data)
        out = cursor.fetchall()
        return out
    elif not door.is_ring:  # If door isn't ring, it probably doesn't have the line parallel with the rooms it joins
        #print 'Door is not ring'
        selectsql = 'select door, st_x(st_centroid(door)) x, st_y(st_centroid(door)) y ' \
                    + 'from (' + dumpdoorlinesegs + ') d where st_isvalid(door) order by st_length(door) desc limit 1'
        insertsql = 'select ST_AsGeoJSON(st_buffer(st_setsrid(st_affine(door, 1, 0, 0, 1, %s - x, %s - y), 900913), ' \
                    + "%s, 'endcap=square join=bevel')) from (" + selectsql + ') t'
        data = (door.center_x, door.center_y, walkable_buffer, tuple(door.door_ids))
        cursor.execute(insertsql, data)
        out = cursor.fetchall()
        return out
    else:
        dumproomlinesegs = 'select st_exteriorring(wkb_geometry) room from floorplan ' \
                           + "where st_geometrytype(wkb_geometry) ilike 'st_polygon' and ogc_fid in %s"
        sql = 'select st_distance(st_pointn(door, 1), room) + st_distance(st_pointn(door, 2), room) maxdist, ' \
              + 'door from (' + dumpdoorlinesegs + ') d ' \
              + 'cross join (' + dumproomlinesegs + ') r where st_isvalid(door) and st_isvalid(room) ' \
              + 'order by maxdist asc nulls last limit 1'

        insertsql = 'select ST_AsGeoJSON(st_buffer(st_makeline(st_pointfromtext(%s, 900913), ' \
                    + "st_centroid(door)), %s, 'endcap=square join=bevel')) from (" + sql + ") t"
        #print insertsql

        data = (door.center_wkt(), walkable_buffer, tuple(door.door_ids), tuple(door.room_ids))
        #print str(data)

        cursor.execute(insertsql, data)
        out = cursor.fetchall()
        return out


# Get all room polygons
def getRoomList():
    dbCursor.execute(
        "SELECT ST_AsGeoJSON(wkb_geometry) FROM floorplan WHERE z=" + str(floorValue) + " AND building_id = " + str(
            buildingId) + " AND (GeometryType(wkb_geometry) = 'POLYGON') AND (layer_name = 'ROM' OR layer_name = 'ROM_ROBUST')")
    rooms = dbCursor.fetchall()

    myList = []

    for room in rooms:
        jsonStr = json.loads(room[0])
        myList.append(jsonStr['coordinates'][0])

    return myList


# Gets already existing paths from the database
# Returns a list of lines
def getExistingPaths():
    existingLines = set()
    # Database query
    # Fetch all paths (ways)
    dbCursor.execute(
        "SELECT ST_AsGeoJSON(geom_way), target, source, id FROM way WHERE (z1 = " + str(floorValue) + " OR z2= " + str(
            floorValue) + ") AND (building1_id = " + str(buildingId) + " OR building2_id = " + str(
            buildingId) + ") AND (GeometryType(geom_way) = 'LINESTRING')")
    dbLines = dbCursor.fetchall()

    # JSON parsing
    for line in dbLines:

        # Add the way source and target to a set. Planned usage: dont make paths to it if there exists paths to it already
        # A set makes sure its unique too.
        poiIds.add(line[1])
        poiIds.add(line[2])

        lineId = line[3]

        jsonLineStr = json.loads(line[0])
        lineStr = jsonLineStr['coordinates']
        # Lets go through all coordinates and create lines to each

        previousCoordinate = None
        for coordinate in lineStr:

            if previousCoordinate is not None:
                # Create points for the previous coordinate and for the new one
                point1 = Point(importLon(previousCoordinate[0]), importLat(previousCoordinate[1]))
                point2 = Point(importLon(coordinate[0]), importLat(coordinate[1]))

                # Create a new line
                newLine = Line(point1, point2)
                newLine.dbId = lineId
                existingLines.add(newLine)

            # Create a previous coordinate so we can make lines between coordinates next to each other.
            previousCoordinate = coordinate

    return existingLines


# Function scanning through our image, trying to find squares without black points inside
def isValidTilePos(x, y):
    # we check if there's any black pixels around our x,y-position
    if (x + (distanceToCheck + (tileSize + 1))) < screenWidth and (y + (distanceToCheck + (tileSize + 1))) < screenHeight and x - (distanceToCheck + (tileSize + 1)) > 0 and y - (distanceToCheck + (tileSize + 1)) > 0:
        for px in xrange(x - (distanceToCheck + (tileSize + 1)), x + (distanceToCheck + (tileSize + 1))):
            for py in xrange(y - (distanceToCheck + (tileSize + 1)), y + (distanceToCheck + (tileSize + 1))):
                if screenBuffer[px][py] == 0:
                    return False

        pygame.draw.rect(allScreen, pygame.color.Color("grey"), (x - int(tileSize / 2), y - int(tileSize / 2), tileSize, tileSize), 1)
        tile = Tile(x, y)
        tiles.add(tile)

        # Found a valid tile
        return True

    # If the tests aren't valid, we don't make a tile :(
    return False


# Help function that runs A*, minimizes the path and later on draws it
def runAStar(start, goal):
    path = AStar(start, goal)

    if path:
        # Add all paths into one big array
        allPaths.update(path)

        # Use the A* path result to create paths consisting of fewer points.
        minimizePath(path)

        # Just a little function to draw the A* result visually.
        drawAStar(path)

    else:
        print "Found no available path for goal at " + str(goal.x) + " / " + str(goal.y)


# This function draws A*
def drawAStar(path):
    global tileSize

    r = random.randint(0, 255)
    g = random.randint(0, 255)
    b = random.randint(0, 255)

    for tile in path:
        tile.timesUsed += 1

        # Random color
        pygame.draw.rect(pathScreen, (r, g, b),
                         (tile.x - int(tileSize / 2), tile.y - int(tileSize / 2), tileSize, tileSize))

        if tile.x == start.x and tile.y == start.y:
            pygame.draw.rect(pathScreen, pygame.color.Color("green"),
                             (tile.x - int(tileSize / 2), tile.y - int(tileSize / 2), tileSize, tileSize))

        if tile.x == goalDest.x and tile.y == goalDest.y:
            pygame.draw.rect(pathScreen, pygame.color.Color("red"),
                             (tile.x - int(tileSize / 2), tile.y - int(tileSize / 2), tileSize, tileSize))


# Function for sorting our min-heap in A*
def sortByWallCost(tile):
    return tile.wallCosts


# This function is called several times to spread the wall costs over all our valid tiles.
def calculateWallCosts():
    z = 3
    for tile in tiles:
        neighbors = tile.neighbors
        if neighbors:
            highestWallCost = max(neighbors, key=sortByWallCost).wallCosts
            if highestWallCost / z > tile.wallCosts:
                tile.wallCosts = highestWallCost / z


# Bresenham's line algorithm
def getLinePoints(x0, y0, x1, y1):
    points = []
    dx = abs(x1 - x0)
    dy = abs(y1 - y0)
    if x0 < x1:
        sx = 1
    else:
        sx = -1
    if y0 < y1:
        sy = 1
    else:
        sy = -1

    err = dx - dy

    while True:
        newPoint = Point(x0, y0)
        points.append(newPoint)
        if x0 == x1 and y0 == y1:
            break
        e2 = 2 * err
        if e2 > -dy:
            err = err - dy
            x0 += sx

        if e2 < dx:
            err = err + dx
            y0 += sy

    return points


# Check if a line crosses a black pixel
def isValidLine(x0, y0, x1, y1):
    screenBuf = pygame.surfarray.pixels2d(allScreen)  # Our buffer of the image
    pts = getLinePoints(x0, y0, x1, y1)
    for point in pts:
        if screenBuf[point.x][point.y] == 0:
            del screenBuf
            return False
    del screenBuf  # Delete the buffer before we continue
    return True


# Takes the output from A* and tries to generate straight lines from the goal to the start.
# It tries to use fewer tiles in total in its path, and wants to connect to the first already existing line in sight (not passing through a wall or similar)
def minimizePath(path):
    # Make sure we have a valid path, otherwise it makes no sense to continue.
    if path:

        newPath = path
        newPath.reverse()  # Reversing the paths, we want to start from the goal and head towards the start,
        # this way we can possibly avoid some extra paths if we find already existing paths before hitting the start.

        stop = None
        validTile = None
        recentlyAddedLines = set()

        for tile in newPath:
            stop = tile
            validTile = tile
            break

        # Check if we have any existing lines
        if lines:

            # Go through every point in our reversed A* result
            for tile in newPath:
                # Skip tile if its the same as our last stop
                if tile == stop:
                    continue

                # If we find a valid, we can try to skip to the next tile to minimize amount of nodes created.
                if isValidLine(stop.x, stop.y, tile.x, tile.y) and isValidLine(stop.x, stop.y, newPath[-1].x, newPath[-1].y) == False:
                    validTile = tile
                else:
                    closestValidLine = None
                    closestValidPoint = None
                    closestDist = None
                    for line in lines:

                        # Dont want to connect to something we just added, as we will just connect to ourselves most likely.
                        if line in recentlyAddedLines:
                            continue

                        allPoints = getLinePoints(line.point1.x, line.point1.y, line.point2.x, line.point2.y)
                        pointsToCheck = []
                        for x in xrange(0, len(allPoints), distanceBetweenPoints):
                            if x < len(allPoints):
                                pointsToCheck.append(allPoints[x])

                        # Add the last point too, the end of the line is a bit more important than the rest of the points.
                        pointsToCheck.append(allPoints[-1])

                        for pt in pointsToCheck:
                            if isValidLine(stop.x, stop.y, pt.x, pt.y):
                                if closestValidPoint is None:
                                    if pt.x == stop.x and pt.y == stop.y and stop.origLon and stop.origLat:
                                        pt.origLon = stop.origLon
                                        pt.origLat = stop.origLat

                                    if stop.isPoi:
                                        pt.isPoi = True

                                    if stop.dbId:
                                        pt.dbId = stop.dbId

                                    closestValidPoint = pt

                                    closestValidLine = line
                                    closestDist = linear_distance(stop.x, stop.y, pt.x, pt.y)
                                else:
                                    distToCurPoint = linear_distance(stop.x, stop.y, pt.x, pt.y)
                                    if distToCurPoint < closestDist:
                                        closestValidLine = line
                                        closestValidPoint = pt
                                        closestDist = distToCurPoint

                    # This is the last point
                    if closestValidPoint:
                        # Since we're adding a new line we want to modify the already existing lines, so we don't have overlapping lines

                        # Create copies of the points to re-use them in new lines
                        p1 = copy.deepcopy(closestValidLine.point1)
                        p2 = copy.deepcopy(closestValidLine.point2)

                        modifyLine = Line(closestValidPoint, p2)
                        lines.add(modifyLine)
                        modifiedLines.add(modifyLine)

                        closestValidLine.point2 = closestValidPoint
                        # We add modified lines to our modifiedLines list, so we can later find what lines have been modified and change them in the database.
                        modifiedLines.add(closestValidLine)

                        # Also add the new connecting line.

                        stopPoint = Point(stop.x, stop.y)
                        if stop.origLon:
                            stopPoint.origLon = stop.origLon
                            stopPoint.origLat = stop.origLat

                        if stop.isPoi:
                            stopPoint.isPoi = True

                        if stop.dbId:
                            stopPoint.dbId = stop.dbId

                        # Adding a new line.
                        newLine = Line(stopPoint, closestValidPoint)
                        modifiedLines.add(newLine)

                        if stopPoint.x == closestValidPoint.x and stopPoint.y == closestValidPoint.y:
                            pass

                        lines.add(newLine)
                        return

                    if validTile and stop != validTile:
                        stopPoint = Point(stop.x, stop.y)
                        if stop.origLon:
                            stopPoint.origLon = stop.origLon
                            stopPoint.origLat = stop.origLat

                        if stop.isPoi:
                            stopPoint.isPoi = True

                        if stop.dbId:
                            stopPoint.dbId = stop.dbId

                        validPoint = Point(validTile.x, validTile.y)
                        if validTile.origLon and validTile.origLat:
                            validPoint.origLon = validTile.origLon
                            validPoint.origLat = validTile.origLat

                        if validTile.isPoi:
                            validPoint.isPoi = True

                        if validTile.dbId:
                            validPoint.dbId = validTile.dbId

                        newLine = Line(stopPoint, validPoint)

                        if stopPoint.x == validPoint.x and stopPoint.y == validPoint.y:
                            pass

                        lines.add(newLine)
                        modifiedLines.add(newLine)
                        recentlyAddedLines.add(newLine)
                        stop = validTile
                        validTile = None

            # This should be the last tile in the path, so lets just add it and return.
            # The process should be almost equal as above
            if validTile:
                if stop.x == validTile.x and stop.y == validTile.y:
                    # No point making a line to the same point.
                    pass
                    return

                stopPoint = Point(stop.x, stop.y)
                if stop.origLon:
                    stopPoint.origLon = stop.origLon
                    stopPoint.origLat = stop.origLat

                if stop.isPoi:
                    stopPoint.isPoi = True

                if stop.dbId:
                    stopPoint.dbId = stop.dbId

                validPoint = Point(validTile.x, validTile.y)
                if validTile.origLon and validTile.origLat:
                    validPoint.origLon = validTile.origLon
                    validPoint.origLat = validTile.origLat

                if validTile.isPoi:
                    validPoint.isPoi = True

                if validTile.dbId:
                    validPoint.dbId = validTile.dbId

                newLine = Line(stopPoint, validPoint)
                if stopPoint.x == validPoint.x and stopPoint.y == validPoint.y:
                    pass
                lines.add(newLine)
                modifiedLines.add(newLine)  # New line modified
                recentlyAddedLines.add(newLine)
                return


        else:  # No lines found
            # Since we have no lines, we create lines ourselves, without checking for lines :)
            for tile in newPath:
                # Skip tile if its the same as our last stop
                if tile == stop:
                    continue

                if isValidLine(stop.x, stop.y, tile.x, tile.y):
                    validTile = tile
                else:
                    if validTile and stop != validTile:
                    #pygame.draw.line(allScreen, (80,80,80), (stop.x, stop.y), (validTile.x, validTile.y), 2)

                        stopPoint = Point(stop.x, stop.y)
                        if stop.origLon:
                            stopPoint.origLon = stop.origLon
                            stopPoint.origLat = stop.origLat

                        if stop.isPoi:
                            stopPoint.isPoi = True

                        if stop.dbId:
                            stopPoint.dbId = stop.dbId

                        validPoint = Point(validTile.x, validTile.y)
                        if validTile.origLon and validTile.origLat:
                            validPoint.origLon = validTile.origLon
                            validPoint.origLat = validTile.origLat
                            print "Set origLon/Lat"

                        if validTile.isPoi:
                            validPoint.isPoi = True

                        if validTile.dbId:
                            validPoint.dbId = validTile.dbId

                        newLine = Line(stopPoint, validPoint)

                        if stopPoint.x == validPoint.x and stopPoint.y == validPoint.y:
                            pass

                        lines.add(newLine)
                        modifiedLines.add(newLine)
                        stop = validTile
                        recentlyAddedLines.add(newLine)


# This is a proof-of-concept function, importing paths as its own layer in qgis.
# It's currently not very useful, but it shows it can be done.
# At the moment we instead import directly to the database.
def exportToQGIS():
    try:
        import PyQt4.QtCore
        import PyQt4.QtGui
        from qgis.core import QgsMapLayerRegistry
        from qgis.core import QgsGeometry
        from qgis.core import QgsPoint
        from qgis.core import QgsVectorLayer
        from qgis.core import QgsFeature

        import qgis.gui
        import qgis.utils

        print "Export to QGIS running."

        # Lets create our own vector layer and put stuff on it :)
        vectorlayer = QgsVectorLayer("Point", "temporary_points", "memory")
        dataprovider = vectorlayer.dataProvider()
        for tile in allPaths:
            fet = QgsFeature()
            fet.setGeometry(QgsGeometry.fromPoint(
                QgsPoint(lowestLon + (tile.x / magnification), lowestLat + (tile.y / magnification))))
            dataprovider.addFeatures([fet])

        vectorlayer.updateExtents()
        print QgsMapLayerRegistry.instance()
        QgsMapLayerRegistry.instance().addMapLayers([vectorlayer])  # Our memory layer...

    except ImportError:
        print "Cannot export to qgis as the plugin couldn't import qgis modules - you're probably not running through the plugin through qgis."


#####
# Functions related to exporting to the database
#####


# Check if a poi id exists in the database.
def hasNodeId(nodeId):
    global dbCursor
    queryStr = "SELECT id FROM node WHERE id = %s"
    data = (nodeId,)

    dbCursor.execute(queryStr, data)

    nodes = dbCursor.fetchall()

    # If we have any results, we return true
    if len(nodes) > 0:
        return True

    return False


# Function to check if a poi-id exists as a path in the way-table.
def nodeHasWay(nodeId):
    global dbCursor
    queryStr = "SELECT id FROM way WHERE (way.building1_id = %s OR way.building2_id = %s) AND (way.target = %s OR way.source = %s) LIMIT 1"
    data = (buildingId, buildingId, nodeId, nodeId)

    dbCursor.execute(queryStr, data)

    result = dbCursor.fetchall()

    # If we have any results, we return true
    if len(result) > 0:
        return True

    return False


# Inserts a new point/node into the poi-table in the database
# Returns false if something goes wrong, otherwise returns the ID
def insertNode():
    global dbCursor
    global dbConn

    queryStr = "INSERT INTO node DEFAULT VALUES returning id"
    dbCursor.execute(queryStr)

    returnId = dbCursor.fetchone()[0]
    dbConn.commit()

    if returnId:
        return returnId

    return None


def modifyPoi(poiId, x, y, z, building, campus):
    # We keep this in case we later want to modify POIs in the database.
    pass


# Check if a way id exists in the database.
def hasWayId(wayId):
    global dbCursor
    queryStr = "SELECT id FROM way WHERE id = %s LIMIT 1"
    data = (wayId,)

    dbCursor.execute(queryStr, data)

    ways = dbCursor.fetchall()

    # If we have any results, we return true
    if len(ways) > 0:
        return True

    return False


# Inserting a path into the database (way table)
def insertWay(x1, y1, z1, x2, y2, z2, buildingId1, buildingId2, source, target, campus_id):
    global dbCursor
    global dbConn

    queryStr = "INSERT INTO way (x1,y1,z1, x2,y2,z2, geom_way, building1_id, building2_id, cost, reverse_cost, source, target, campus_id ) VALUES (%s, %s, %s, %s, %s, %s, ST_GeomFromText(%s, 900913), %s, %s, %s, %s, %s, %s, %s) RETURNING id;";
    geometryStr = "LINESTRING(" + str(x1) + " " + str(y1) + "," + str(x2) + " " + str(y2) + ")"

    data = (x1, y1, z1, x2, y2, z2, geometryStr, buildingId1, buildingId2, 10, 10, source, target, campus_id)
    dbCursor.execute(queryStr, data)

    returnId = dbCursor.fetchone()[0]
    dbConn.commit()

    if returnId:
        return returnId

    return None


def modifyWay(wayId, x1, y1, z1, x2, y2, z2, buildingId1, buildingId2, source, target, campus_id):
    global dbCursor
    global dbConn
    queryStr = "UPDATE way SET x1 = %s, y1 = %s, z1 = %s, x2 = %s,y2 = %s, z2 = %s, geom_way = ST_GeomFromText(%s, 900913), building1_id = %s, building2_id = %s, cost = %s, reverse_cost = %s, source = %s, target = %s, campus_id = %s WHERE id = %s"
    geometryStr = "LINESTRING(" + str(x1) + " " + str(y1) + "," + str(x2) + " " + str(y2) + ")"

    # TODO: Fix cost, it's not incredibly important that it's correct, but just letting this stay here.
    data = (x1, y1, z1, x2, y2, z2, geometryStr, buildingId1, buildingId2, 10, 10, source, target, campus_id, wayId)

    dbCursor.execute(queryStr, data)
    print "Updated wayId " + str(wayId)


# Main function starting up everything
def main():
    # Globaling all the variables that will need to be changed at some point.
    global lowestLon
    global lowestLat
    global doorList
    global myList
    global maxWindowHeight
    global maxWindowWidth
    global screenWidth
    global screenHeight
    global screenBuffer
    global allScreen
    global pathScreen
    global start
    global goal
    global goalDest
    global hasPlacedStart
    global hasPlacedEnd
    global doorPolys
    global allPaths
    global screen
    global lines
    global screenBuf
    global dbConn
    global dbCursor
    global poiIds

    # Fetch all the rooms
    myList = getRoomList()

    for polygon in myList:
        for lonLat in polygon:
            curLon = lonLat[0]
            curLat = lonLat[1]

            if lowestLon is None or lowestLon > curLon:
                lowestLon = curLon

            if lowestLat is None or lowestLat > curLat:
                lowestLat = curLat

    #print "Lon: " + str(lowestLon) + " Lat: " + str(lowestLat)

    # getDoors fetches all doors and adds them to a previously created doorPolys list.
    getDoors()

    maxWindowHeight = None
    maxWindowWidth = None

    # Lets find the proper window size before we draw the polygons
    for polygon in myList:
        for lonLat in polygon:
            curLon = (lonLat[0] - lowestLon) * magnification
            curLat = (lonLat[1] - lowestLat) * magnification

            if maxWindowWidth is None or maxWindowWidth < curLon:
                maxWindowWidth = curLon

            if maxWindowHeight is None or maxWindowHeight < curLat:
                maxWindowHeight = curLat

    maxWindowHeight = int(math.ceil(maxWindowHeight))
    maxWindowWidth = int(math.ceil(maxWindowWidth))

    print "Window size: " + str(maxWindowHeight) + " / " + str(maxWindowWidth)

    print "Importer running"
    pygame.init()
    size = width, height = maxWindowWidth, maxWindowHeight
    pygame.display.set_caption('Walkthrough importer')
    screen = pygame.display.set_mode(ScrSize)
    allScreen = pygame.Surface((maxWindowWidth, maxWindowHeight))
    popupScreen = pygame.Surface((maxWindowWidth, maxWindowHeight))
    popupScreen.set_colorkey((0, 0, 0))
    popupScreen.set_alpha(208)
    pathScreen = pygame.Surface((maxWindowWidth, maxWindowHeight))

    bg = pygame.Surface(ScrSize).convert()
    initRect = pygame.Rect(screen.get_rect())

    # The scrollbar being used is ezscroll, which can be found at https://code.google.com/p/ezscroll/

    sp = ScrollPane(
        allScreen.get_size(),
        initRect,
        allScreen,
        bg,
        [W, N],
        2,
        True,
        scrollBarSize)
    sp.draw(bg)
    screen.blit(bg, Origin)
    pygame.display.flip()

    pathScreen.set_colorkey((0, 0, 0))
    pathScreen.set_alpha(208)
    for _ in xrange(maxWindowWidth):
        newTiles.append([0] * maxWindowHeight)

    # Now lets draw the polygons.
    for polygon in myList:
        newList = []
        for lonLat in polygon:
            item = []
            curLon = (lonLat[0] - lowestLon) * magnification
            curLat = (lonLat[1] - lowestLat) * magnification
            item.append(curLon)
            item.append(curLat)
            newList.append(item)

        # Draw the polygons (Rooms)
        pygame.draw.polygon(allScreen, pygame.color.Color("white"), newList, 0);

    for polygon in doorPolys:
        newList = []
        for lonLat in polygon:
            item = []
            curLon = (lonLat[0] - lowestLon) * magnification
            curLat = (lonLat[1] - lowestLat) * magnification
            item.append(curLon)
            item.append(curLat)
            newList.append(item)

        pygame.draw.polygon(allScreen, pygame.color.Color("green"), newList, 0)

    if manualDoors:
        print "Will be manually drawing doors now, click with the mouse to create a door (marked with blue). Press D to finish."
        drawingDoors = True
        while drawingDoors:
            for event in pygame.event.get():
                if event.type == pygame.KEYDOWN:
                # treat the Q and escape keys as quit commands
                    if event.key == pygame.K_d:
                        drawingDoors = False

                if event.type == pygame.QUIT: pygame.quit()
                if event.type == MOUSEBUTTONDOWN:
                    mousex, mousey = pygame.mouse.get_pos()
                    pygame.draw.circle(allScreen, pygame.color.Color("light blue"), (mousex, mousey), doorExpandRadius,
                                       0)

            # Update the display
            pygame.display.update()

    # Get all the POIs we should visit
    poiList = getPOIList()

    # Fetch all exisiting paths...
    existingPaths = getExistingPaths()

    #existingPaths = set() # Debug to remove all existing paths
    #poiIds = set() # Debug to remove all existing paths

    # Lets get all existing paths to connect to them
    exPaths = copy.deepcopy(existingPaths)
    lines.update(exPaths)

    screenWidth = maxWindowWidth
    screenHeight = maxWindowHeight

    print "Start isValidTilePos"

    screenBuffer = pygame.surfarray.pixels2d(allScreen)  # fast
    #screenBuffer = pygame.PixelArray(allScreen)  # faster, but uses more memory

    print "Screenbuffer has been set!"


    def runValidTest():
        for i in xrange(0, screenWidth, (tileSize + 1)):
            for j in xrange(0, screenHeight, (tileSize + 1)):
                isValidTilePos(i, j)


    runValidTest()
    # for profiling
    #cProfile.run("runValidTest()")

    # Remove the screenBuffer
    del screenBuffer

    print "End isValidTilePos"

    randomPoints = random.sample(tiles, 2)
    start = randomPoints[0]
    goal = randomPoints[1]

    pygame.draw.rect(allScreen, pygame.color.Color("green"),
                     (start.x - int(tileSize / 2), start.y - int(tileSize / 2), tileSize, tileSize))

    pygame.draw.rect(allScreen, pygame.color.Color("red"),
                     (goal.x - int(tileSize / 2), goal.y - int(tileSize / 2), tileSize, tileSize))

    print "Done with making tile-board, lets run AStar next"

    # Update the screen before astar
    pygame.display.update()

    # Lets calculate AStar
    for tile in tiles:
        tile.neighbors = getNeighbours(tile)
        tile.wallCosts = (8 - len(
            tile.neighbors)) * wallCostAmount # Lets set the initial wallcost here. 8 is the total amount of neighbors, we want to penalize tiles with fewers neighbors.

    # Lets calculate wall costs, has to be done after neighbors and initial wallcosts have been set.
    for i in xrange(4):
        calculateWallCosts()

    splitTime = time.clock()
    if runSearch:
        path = None

        splitTime = time.clock()

        goals = []

        print poiIds

        for poi in poiList:
            if poi.dbId in poiIds:
                # Skip the POI as a goal since it already has a path to it
                print "Already have a path to " + str(poi.dbId)
                continue

            path = None

            # Lets make a goal poi
            tile = Tile((int)(poi.x - lowestLon) * magnification, int(poi.y - lowestLat) * magnification)
            tile.isPoi = True

            # If we have a POI from the database, then we save the dbId for the tile
            # That way we can later connect paths to this ID
            if poi.dbId:
                tile.dbId = poi.dbId

            # Show it in the UI

            closestTile = None
            closestDist = None
            for t in tiles:
                curDist = distance(tile.x, tile.y, t.x, t.y)
                if closestTile is None or closestDist is None or curDist < closestDist:
                    closestTile = t
                    closestDist = curDist

                # Lets break out of the tiles loop if we find a tile closer than the tileSize.
                # Might as well save some time if we can.
                if curDist < tileSize:
                    break

            pygame.draw.circle(allScreen, pygame.color.Color("orange"), (closestTile.x, closestTile.y), tileSize * 2, 3)
            if closestTile:
                closestTile.origLon = poi.x
                closestTile.origLat = poi.y
                closestTile.isPoi = True
                closestTile.dbId = poi.dbId
                goals.append(closestTile)

        random.shuffle(goals)

        for goalDest in goals:

            print " ----------- new goalDest " + str(len(lines))

            if lines:
                closestLine = None
                closestPoint = None
                for line in lines:
                    pts = getLinePoints(line.point1.x, line.point1.y, line.point2.x, line.point2.y)
                    for point in pts:

                        if closestPoint is None:
                            closestPoint = point
                            closestLine = line

                        if linear_distance(point.x, point.y, goalDest.x, goalDest.y) < linear_distance(closestPoint.x,
                                                                                                       closestPoint.y,
                                                                                                       goalDest.x,
                                                                                                       goalDest.y):
                            closestPoint = point
                            closestLine = line

                # Find closest tile to point
                for t in tiles:
                    if closestPoint and (
                        linear_distance(t.x, t.y, closestPoint.x, closestPoint.y) < linear_distance(start.x, start.y,
                                                                                                    closestPoint.x,
                                                                                                    closestPoint.y)):
                        start = t

            else:
                start = random.sample(goals, 1)[0]

            path = None
            runAStar(start, goalDest)

            # If we just want one path returned, instead of all POIs.
            if generateOnePath:
                break
    endTime = time.clock()

    print "Start time: " + str(startTime) + " Split: " + str((splitTime - startTime)) + " End: " + str(
        (endTime - startTime))

    print "Tiles size: " + str(len(tiles))
    print "Start: " + str(start.x) + " / " + str(start.y) + " End: " + str(goal.x) + " / " + str(goal.y)


    # Lets try to export to qgis.
    exportToQGIS()

    # Drawing all lines close to black.
    for lineToDraw in lines:
        r = g = b = 1

        pygame.draw.line(allScreen, (r, g, b), (lineToDraw.point1.x, lineToDraw.point1.y),
                         (lineToDraw.point2.x, lineToDraw.point2.y), 3)

    # Drawing existing paths in a different color to separate them from newly generated ones
    for lineToDraw in existingPaths:
        pygame.draw.line(allScreen, (210, 0, 0), (lineToDraw.point1.x, lineToDraw.point1.y),
                         (lineToDraw.point2.x, lineToDraw.point2.y), 3)

    # Lets draw to the screen to take a screenshot
    screen.fill((0, 0, 0))
    screen.blit(allScreen, (0, 0))
    #screen.blit(pathScreen, (0,0))
    pygame.display.flip()
    pygame.image.save(allScreen,
                      "build-" + str(buildingId) + "-" + str(floorValue) + " .jpg")

    popupPositionX = 0
    popupPositionY = 0
    font = pygame.font.Font(None, 16)
    txt = font.render("", True, (255, 255, 255))

    ##### RESULT PRINTING

    roomAmount = len(myList)

    # Sets are unique so it should get all points based on the lines
    pointAmount = set()
    for li in lines:

        if li.point1.x == li.point2.x and li.point1.y == li.point2.y:
            pass

        pointAmount.add(hash(str(li.point1.x) + " " + str(li.point1.y)))
        pointAmount.add(hash(str(li.point2.x) + " " + str(li.point2.y)))
        print li.toString()

    print "Building id, floor, Rooms, POIs, Paths, Missing, Time"
    print str(buildingId) + " & " + str(floorValue) + " & " + str(roomAmount) + " & " + str(len(goals)) + " & " + str(
        len(lines)) + " & 0 & " + str(("%.2f" % (endTime - startTime))) + " \\\\ \hline"
    with open("results.txt", "a") as resultFile:
        resultFile.write(
            str(buildingId) + " & " + str(floorValue) + " & " + str(roomAmount) + " & " + str(len(goals)) + " & " + str(
                len(lines)) + " & 0 & " + str(("%.2f" % (endTime - startTime))) + " \\\\ \hline\n")
        resultFile.close()


    #
    if shouldExitOnCompletion > 0:
        sys.exit()

    # Loop doing handling drawing all graphics and mouse/key-presses.
    while 1:

        # Just drawing all lines already generated onto allScreen
        for lineToDraw in lines:
            #        r = random.randint(0,255)
            #        g = random.randint(0,255)
            #        b = random.randint(0,255)
            r = g = b = 1

            pygame.draw.line(allScreen, (r, g, b), (lineToDraw.point1.x, lineToDraw.point1.y),
                             (lineToDraw.point2.x, lineToDraw.point2.y), 3)


        # Reseting the screen
        screen.fill((0, 0, 0))

        # Drawing the allScreen surface onto the screen
        screen.blit(allScreen, (0, 0))

        # Get the keys that are currently being held down
        keysDown = pygame.key.get_pressed()
        if keysDown[pygame.K_p]:
            # show the path.
            screen.blit(pathScreen, (0, 0))

        # Event handler
        for event in pygame.event.get():
            sp.update(event)

            # Mouse movement
            if event.type == pygame.MOUSEMOTION:
                mousex, mousey = pygame.mouse.get_pos()
                closestTile = None
                closestDist = None
                for t in tiles:
                    curDist = distance(mousex, mousey, t.x, t.y)
                    if closestTile is None or closestDist is None or curDist < closestDist:
                        closestTile = t
                        closestDist = curDist

                txtStr = str(closestTile.x) + "/" + str(closestTile.y) + " wallcost: " + str(closestTile.wallCosts)
                txt = font.render(txtStr, True, (255, 255, 255))
                screen.blit(txt, (scrollBarSize + 1, scrollBarSize + 1))

            # Key presses
            if event.type == pygame.KEYDOWN:
            # treat the Q and escape keys as quit commands
                if event.key == pygame.K_q:
                    pygame.quit()
                    return
                elif event.key == pygame.K_ESCAPE:
                    pygame.quit()
                    return
                elif event.key == pygame.K_p:
                    screen.blit(pathScreen, (0, 0))
                elif event.key == pygame.K_u:
                    print "Checking ids: "
                    for line in lines:
                        if line.point1.dbId:
                            print "Found (1) dbId " + str(line.point1.dbId)

                            for poi in poiList:
                                if line.point1.dbId == poi.dbId:
                                    print "Found matching poi 1 to dbId " + str(poi.dbId)

                        if line.point2.dbId:
                            print "Found (2) dbId " + str(line.point2.dbId)
                            for poi in poiList:
                                if line.point2.dbId == poi.dbId:
                                    print "Found matching poi 2 to dbId " + str(poi.dbId)

                elif event.key == pygame.K_i:

                    # Inserting back into the database.
                    for line in lines:
                        if line in modifiedLines: # Check if the line is modified, if it is, then insert it

                            # Set up some of the coordinates we're going to use.
                            # Need to convert from local to proper longitude and latitude
                            p1origLon = exportLon(line.point1.x)
                            p1origLat = exportLat(line.point1.y)

                            p2origLon = exportLon(line.point2.x)
                            p2origLat = exportLat(line.point2.y)

                            # We store the original lon/lats so we don't mess up the database.
                            if line.point1.origLon and line.point1.origLat:
                                p1origLon = line.point1.origLon
                                p1origLat = line.point1.origLat

                            if line.point2.origLon and line.point2.origLat:
                                p2origLon = line.point2.origLon
                                p2origLat = line.point2.origLat

                            if line.point1.dbId and hasNodeId(line.point1.dbId): # and line.point1 in modifiedPoints:
                                # We already have a dbId, and it is modified (update the db-information for that POI
                                modifyPoi(line.point1.dbId, p1origLon, p1origLat, floorValue, buildingId, campusId)
                            else:
                                # We dont have this point in the database
                                # Add the point and then set the points dbId to this new point.
                                    line.point1.dbId = insertNode()

                            # Same code for the second point
                            if line.point2.dbId and hasNodeId(line.point2.dbId): # and line.point1 in modifiedPoints:
                                # We already have a dbId, and it is modified (update the db-information for that POI
                                modifyPoi(line.point2.dbId, p2origLon, p2origLat, floorValue, buildingId, campusId)
                            else:
                                # We don't have this point in the database
                                # Add the point and then set the points dbId to this new point.
                                    line.point2.dbId = insertNode()

                            if line.dbId and hasWayId(line.dbId):
                                # If we reach here, we have to modify the line in the database with the new points etc.
                                modifyWay(line.dbId, p1origLon, p1origLat, floorValue, p2origLon, p2origLat, floorValue,
                                          buildingId, buildingId, line.point1.dbId, line.point2.dbId, campusId)
                            else:
                                insertId = insertWay(p1origLon, p1origLat, floorValue, p2origLon, p2origLat, floorValue,
                                                     buildingId, buildingId, line.point1.dbId, line.point2.dbId,
                                                     campusId)
                                line.dbId = insertId

                                print "Added new id to db: " + str(insertId)

                    dbConn.commit()

            if event.type == pygame.QUIT:
                pygame.quit()
                return

            # Mouse actions, can be used to create new paths or to display debug information or similar.
            if event.type == MOUSEBUTTONDOWN:
                mousex, mousey = pygame.mouse.get_pos()

                # Should not generate paths if we click the scrollbars :)
                if (mousex <= scrollBarSize or mousey <= scrollBarSize):
                    # Clicking the scrollbar, move on!
                    pass
                else:
                    offsets = [0, 0]
                    for test in sp.sprites:
                        offsets[test.axis] = test.get_scrolled()[test.axis]

                    mousex -= scrollBarSize
                    mousex += int(offsets[0])
                    mousey -= scrollBarSize
                    mousey += int(offsets[1])

                    print offsets

                    print str(allScreen.get_at((mousex, mousey)))

                    print "Mouse down: " + str((mousex / magnification) + lowestLon) + "," + str(
                        (mousey / magnification) + lowestLat) + " orig: (" + str(mousex) + "," + str(mousey) + ")"

                    # If we already have lines, we connect to them
                    if lines:
                        closestPoint = None
                        for line in lines:
                            pts = getLinePoints(line.point1.x, line.point1.y, line.point2.x, line.point2.y)
                            for point in pts:

                                if closestPoint is None:
                                    closestPoint = point

                                if linear_distance(point.x, point.y, goalDest.x, goalDest.y) < linear_distance(
                                        closestPoint.x, closestPoint.y, goalDest.x, goalDest.y):
                                    closestPoint = point

                        # Find closest tile to point
                        for t in tiles:
                            if closestPoint and (
                                linear_distance(t.x, t.y, closestPoint.x, closestPoint.y) < linear_distance(start.x,
                                                                                                            start.y,
                                                                                                            closestPoint.x,
                                                                                                            closestPoint.y)):
                                start = t

                        closestStartTile = None
                        startDist = None
                        closestTile = None
                        closestDist = None
                        for t in tiles:
                            curDist = distance(mousex, mousey, t.x, t.y)

                            curStartDist = distance(start.x, start.y, t.x, t.y)

                            if closestStartTile is None or startDist is None or curStartDist < startDist:
                                closestStartTile = t
                                startDist = curStartDist

                            if closestTile is None or closestDist is None or curDist < closestDist:
                                closestTile = t
                                closestDist = curDist

                        mouseClickGoal = closestTile

                        pygame.draw.rect(allScreen, pygame.color.Color("red"), (
                        mouseClickGoal.x - int(tileSize / 2), (mouseClickGoal.y - int(tileSize / 2)), tileSize,
                        tileSize))

                        runAStar(closestStartTile, mouseClickGoal)

                    else: # If we dont have lines, we have to click twice to create paths from one point to another.

                        if hasPlacedStart is None:

                            closestTile = None
                            closestDist = None
                            for t in tiles:
                                curDist = distance(mousex, mousey, t.x, t.y)
                                if closestTile is None or closestDist is None or curDist < closestDist:
                                    closestTile = t
                                    closestDist = curDist

                            mouseClickStart = closestTile
                            tempx = mousex
                            tempy = mousey
                            pygame.draw.rect(allScreen, pygame.color.Color("green"), (
                            mouseClickStart.x - int(tileSize / 2), mouseClickStart.y - int(tileSize / 2), tileSize,
                            tileSize))

                            #pygame.draw.circle(screen, pygame.color.Color("green"), (mousex, mousey), 10, 1)
                            pygame.display.update()
                            hasPlacedStart = True
                        elif hasPlacedEnd is None:

                            print "Line between start point and end point valid: " + str(isValidLine(mousex, mousey, tempx, tempy))

                            closestTile = None
                            closestDist = None
                            for t in tiles:
                                curDist = distance(mousex, mousey, t.x, t.y)
                                if closestTile is None or closestDist is None or curDist < closestDist:
                                    closestTile = t
                                    closestDist = curDist

                            mouseClickGoal = closestTile

                            pygame.draw.rect(allScreen, pygame.color.Color("red"), (
                            mouseClickGoal.x - int(tileSize / 2), (mouseClickGoal.y - int(tileSize / 2)), tileSize,
                            tileSize))

                            runAStar(mouseClickStart, mouseClickGoal)

                            hasPlacedStart = None
                            hasPlacedEnd = None


        # Draw the scrollbars and text.
        for sb in sp.sprites:
            sb.scroll(0.000001)

        # Draw the rest
        changes = sp.draw(bg)
        screen.blit(bg, Origin)
        pygame.display.update(changes)
        screen.blit(txt, (scrollBarSize + 1, scrollBarSize + 1))
        pygame.display.flip()
        pygame.time.delay(100)

# Start the whole system
if __name__ == "__main__":
    main()